local ADDONNAME, MVAnnouncer = ...

-- luacheck: globals LibStub MVAnnouncerDefaults
LibStub("AceAddon-3.0"):NewAddon(MVAnnouncer, ADDONNAME, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

-- load libs
local LSM = LibStub("LibSharedMedia-3.0")
local LSK = LibStub("LibSink-2.0")

-- load locale
local L = LibStub("AceLocale-3.0"):GetLocale(ADDONNAME)

-- WoW APIs
--local _G = _G
local C_Vignettes = C_Vignettes
local C_Vignettes_GetVignetteInfoFromInstanceID = C_Vignettes.GetVignetteInfoFromInstanceID
local GetMapLandmarkInfo = GetMapLandmarkInfo
local LE_MAP_LANDMARK_TYPE_VIGNETTE = LE_MAP_LANDMARK_TYPE_VIGNETTE
local InterfaceOptionsFrame_OpenToCategory, InterfaceOptionsFrameAddOns = InterfaceOptionsFrame_OpenToCategory, InterfaceOptionsFrameAddOns
local GetTime, BetterDate = GetTime, BetterDate

local CHAT_TIMESTAMP_FORMAT = CHAT_TIMESTAMP_FORMAT

-- Lua API
local pairs, type, tonumber, unpack, time = pairs, type, tonumber, unpack, time
local rawget, rawset, wipe = rawget, rawset, wipe
local find, tostring, split, sub, len, format, trim, lower, match = string.find, tostring, string.split, string.sub, string.len, string.format, string.trim, string.lower, string.match

-- GLOBALS: LibStub, MVAnnouncerDefaults, NUM_WORLDMAP_POIS

local defaultSoundName = "Gruntling Horn"
LSM:Register(LSM.MediaType.SOUND, defaultSoundName, "Sound/Events/gruntling_horn_bb.ogg")

local objectIconTextureSize = 480
local OBJECTICONS_texture = "Interface\\MINIMAP\\ObjectIconsAtlas"
local OBJECTICONS_chatTexture = ("|T%s:%%d:%%d:0:0:%d:%d:%%d:%%d:%%d:%%d|t"):format(OBJECTICONS_texture, objectIconTextureSize, objectIconTextureSize)
local OBJECTICONS_coords
do-- OBJECTICONS_coords
    local GetObjectIconTextureCoords = GetObjectIconTextureCoords

    OBJECTICONS_coords = setmetatable({}, {__index = function(t,k)
        local v = { GetObjectIconTextureCoords(k) }
        rawset(t, k, v)

        return v
    end
    })
end

local SPLITKEY = "|"
local sinkTable = setmetatable({}, {__index=function(t,key)
    local v = LSK:Embed({})
    if find(key, SPLITKEY) < 2 then
        v:SetSinkStorage({})
    else
        local objectIcon, name, sinkKey = split(SPLITKEY, key)
        v:SetSinkStorage(MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[sinkKey])
    end

    rawset(t, key, v)

    return v
end})

local function HexToRGBA(hex)
    local rhex, ghex, bhex, ahex = sub(hex, 1, 2), sub(hex, 3, 4), sub(hex, 5, 6), sub(hex, 7, 8)

    if len(ahex) == 2 then
        return tonumber(ghex, 16)/255, tonumber(bhex, 16)/255, tonumber(ahex, 16)/255, tonumber(rhex, 16)/255
    elseif len(bhex) == 2 then
        return tonumber(rhex, 16)/255, tonumber(ghex, 16)/255, tonumber(bhex, 16)/255
    end

    return 1.0, 1.0, 1.0, 1.0
end

local function RGBAToHex(r,g,b,a)
    r = r <= 1 and r >= 0 and r or 0
    g = g <= 1 and g >= 0 and g or 0
    b = b <= 1 and b >= 0 and b or 0
    a = (type(a)=="number") and (a <= 1 and a >= 0 and a) or nil

    if a == nil then
        return format( ("%02x%02x%02x"), r*255, g*255, b*255)
    else
        return format( ("%02x%02x%02x%02x"), a*255, r*255, g*255, b*255)
    end
end

local function GetBooleanValue(input)
    input = trim( lower( tostring(input) ) )
    if input == "true" or input == "1" then
        return true, true
    elseif input == "false" or input == "0" then
        return false, true
    elseif input == "nil" then
        return nil, true
    end

    return nil, false
end

local defaults = {
    global = {
        --[===[@debug@
        vignette_cache = {
            ['**'] = {
                ['**'] = {
                },
            },
        },
        --@end-debug@]===]
        -- landmarkOfInterests = {},
    },
    profile = {
        scan = {
            active = false,
            interval = 3,
            objects = {
            },
        },
        cooldownTime = 10,
        soundCooldown = 2,
        active = true,
        objectIconTypeOfInterest_version = 0,
        objectIconTypeOfInterest = {
            ['**'] = {
                active = true,
                ['**'] = MVAnnouncerDefaults,
            },
        },
        announce_map_vignette = true,
    }
}


function MVAnnouncer:OnInitialize()
    if type(MVAnnouncerDefaults) ~= "table" then
        MVAnnouncerDefaults = {}
    end
    if MVAnnouncerDefaults.seen == nil then
        MVAnnouncerDefaults.seen = false
    end
    if MVAnnouncerDefaults.active == nil then
        MVAnnouncerDefaults.active = true
    end
    if MVAnnouncerDefaults.showIconBefore == nil then
        MVAnnouncerDefaults.showIconBefore = true
    end
    if MVAnnouncerDefaults.showIconAfter == nil then
        MVAnnouncerDefaults.showIconAfter = false
    end
    if MVAnnouncerDefaults.mainTextColour == nil then
        MVAnnouncerDefaults.mainTextColour = "FFFFFFFF"
    end
    if MVAnnouncerDefaults.objectTextColour == nil then
        MVAnnouncerDefaults.objectTextColour = "FFB58900"
    end
    if MVAnnouncerDefaults.iconSize == nil then
        MVAnnouncerDefaults.iconSize = 16
    end
    if MVAnnouncerDefaults.playSound == nil then
        MVAnnouncerDefaults.playSound = true
    end
    if MVAnnouncerDefaults.sound == nil then
        MVAnnouncerDefaults.sound = defaultSoundName
    end
    if type(MVAnnouncerDefaults.sinks) ~= "table" then
        MVAnnouncerDefaults.sinks = {}
    end
    if MVAnnouncerDefaults.sinks['**'] == nil then
        MVAnnouncerDefaults.sinks['**'] = {
            sink20OutputSink = "RaidWarning",
        }

        -- First time we load this, so add a default sink
        MVAnnouncerDefaults.sinks.BuildInDefaultOutputSink = {
            sink20OutputSink = "RaidWarning",
        }
    end

    defaults.profile.objectIconTypeOfInterest['**']['**'] = MVAnnouncerDefaults
    self.db = LibStub("AceDB-3.0"):New("MVAnnouncerDB", defaults, true)
    self.db.RegisterCallback(self, "OnProfileCopied", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnDatabaseReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnNewProfile", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileChanged", "OnDBUpdate")

    self:OnDBUpdate("OnProfileLoaded", self.db, self.db:GetCurrentProfile())

    self:RegisterChatCommand("minimapvignetteannouncer", "CMD")
    self:RegisterChatCommand("mvannouncer", "CMD")
    self:RegisterChatCommand("mva", "CMD")
    if L["CHAT_COMMAND"] ~= "CHAT_COMMAND" then
        self:RegisterChatCommand(L["CHAT_COMMAND"], "CMD")
    end

    self:RegisterOptions()
end

do-- OnDBUpdate
    local OLD_TO_NEW_OBJECTICONS = {
        ["3"] = "4704",
        ["37"] = "4720",
        ["40"] = "4715",
        ["41"] = "4733",
        ["42"] = "4716",
        ["45"] = "4729",
        ["49"] = "49",
        ["5"] = "4703",
    }

    function MVAnnouncer:OnDBUpdate(event, db, profileName)
        --[===[@debug@
        self:Print("OnDBUpdate", event, profileName)
        --@end-debug@]===]
        local profile = (db or self.db).profile

        -- Legion ObjectIcons to ObjectIconsAtlas update
        if profile.objectIconTypeOfInterest_version < 1 then
            local objectIconTypeOfInterest = profile.objectIconTypeOfInterest
            local removeMe = {}

            for oldID, oldData in pairs(objectIconTypeOfInterest) do
                local newID = OLD_TO_NEW_OBJECTICONS[oldID]
                if newID then
                    local newData = objectIconTypeOfInterest[newID]
                    for name, data in pairs(oldData) do
                        local _ = newData[name]

                        if type(data) == "table" then
                            for k, v in pairs(data) do
                                local _ = newData[name][k]
                                newData[name][k] = v
                            end
                        else
                            newData[name] = data
                        end
                    end
                    removeMe[oldID] = true
                end
            end

            for id in pairs(removeMe) do
                objectIconTypeOfInterest[id] = nil
            end

            profile.objectIconTypeOfInterest_version = 1
        end

        self:CleanOptions()
    end
end

function MVAnnouncer:OnEnable()
    self:RegisterEvent("VIGNETTE_ADDED")
    if self.db.profile.announce_map_vignette then
        self:RegisterEvent("WORLD_MAP_UPDATE")
        self:WORLD_MAP_UPDATE()
    end
    self:EnableScan()
end

function MVAnnouncer:OnDisable()
    self:UnregisterEvent("VIGNETTE_ADDED")
    self:UnregisterEvent("WORLD_MAP_UPDATE")
    self:DisableScan()
end

do -- Scanning
    -- Use this only when a vignette on the map with GetNumVignettes() is found
    -- but does not trigger VIGNETTE_ADDED.
    -- enable with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.active = true
    -- disable with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.active = false
    -- add an objectIcon ID: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.objects["<ID>"] = true
    -- remove an objectIcon ID: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.objects["<ID>"] = nil
    -- Set scan interval(seconds) with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.interval = 10
    local C_Vignettes_GetNumVignettes = C_Vignettes.GetNumVignettes
    local C_Vignettes_GetVignetteGUID = C_Vignettes.GetVignetteGUID
    local scanTimer

    function MVAnnouncer:EnableScan()
        if self.db.profile.scan.active then
            self:DisableScan()
            scanTimer = self:ScheduleRepeatingTimer("Scan", self.db.profile.scan.interval)
        else
            self:DisableScan()
        end
    end
    function MVAnnouncer:DisableScan()
        self:CancelTimer(scanTimer)
    end
    function MVAnnouncer:Scan()
        local objects = self.db.profile.scan.objects
        local numVignettes = C_Vignettes_GetNumVignettes()

        for i=1, numVignettes do
            local vignetteGUID = C_Vignettes_GetVignetteGUID(i)
            local ofsX, ofsY, name, objectIcon = C_Vignettes_GetVignetteInfoFromInstanceID(vignetteGUID)
            if objectIcon and objects[tostring(objectIcon)] then
                self:VIGNETTE_ADDED("VIGNETTE_ADDED", vignetteGUID)
            end
        end
    end
end

do-- PlayControlledSoundFile
    local PlaySoundFile = PlaySoundFile
    local cooldown = {}

    function MVAnnouncer:PlayControlledSoundFile(soundFile, channel)
        if ( cooldown[soundFile] or 0 ) <= GetTime() then
            cooldown[soundFile] = GetTime() + self.db.profile.soundCooldown
            PlaySoundFile(soundFile, channel)
        end
    end
end

function MVAnnouncer:Announce(name, objectIcon, isOnMap)
    if not objectIcon then
        objectIcon = L["Unknown Icon"]
    else
        objectIcon = tostring(objectIcon)
    end
    if not name then
        name = L["Unknown vignette"]
    else
        name = tostring(name)
    end

    local objectIconTypeOfInterest = self.db.profile.objectIconTypeOfInterest
    local this_objectIconTypeOfInterest = objectIconTypeOfInterest[objectIcon][name]

    if not this_objectIconTypeOfInterest.seen then
        this_objectIconTypeOfInterest.seen = true
        self:NotifyOptionsChange()
        --for key in pairs(this_objectIconTypeOfInterest.sinks) do
        --    this_objectIconTypeOfInterest.sinks[key].touched = true
        --end
    end

    if objectIconTypeOfInterest[objectIcon].active and this_objectIconTypeOfInterest.active then
        local msg = ""

        local left, right, top, bottom = nil, nil, nil, nil
        if this_objectIconTypeOfInterest.showIconBefore and tonumber(objectIcon) then
            if not left then
                left, right, top, bottom = unpack(OBJECTICONS_coords[objectIcon])
                left, right, top, bottom = left*objectIconTextureSize, right*objectIconTextureSize, top*objectIconTextureSize, bottom*objectIconTextureSize
            end

            msg = msg..format(OBJECTICONS_chatTexture, this_objectIconTypeOfInterest.iconSize, this_objectIconTypeOfInterest.iconSize, left, right, top, bottom)
        end

        local mainTextColour = this_objectIconTypeOfInterest.mainTextColour
        local objectTextColour = this_objectIconTypeOfInterest.objectTextColour

        msg = msg..format( L["|c%sFound |c%s%q|c%s look on your %s!|r"], mainTextColour, objectTextColour, name, mainTextColour, isOnMap and L["map"] or L["minimap"])

        if this_objectIconTypeOfInterest.showIconAfter and tonumber(objectIcon) then
            if not left then
                left, right, top, bottom = unpack(OBJECTICONS_coords[objectIcon])
                left, right, top, bottom = left*objectIconTextureSize, right*objectIconTextureSize, top*objectIconTextureSize, bottom*objectIconTextureSize
            end

            msg = msg..format(OBJECTICONS_chatTexture, this_objectIconTypeOfInterest.iconSize, this_objectIconTypeOfInterest.iconSize, left, right, top, bottom)
        end

        if this_objectIconTypeOfInterest.playSound then
            self:PlayControlledSoundFile(LSM:Fetch(LSM.MediaType.SOUND, this_objectIconTypeOfInterest.sound), "Master")
        end

        for key in pairs(this_objectIconTypeOfInterest.sinks) do
            if CHAT_TIMESTAMP_FORMAT and this_objectIconTypeOfInterest.sinks[key].sink20OutputSink == "ChatFrame" then
                sinkTable[objectIcon..SPLITKEY..name..SPLITKEY..key]:Pour(BetterDate(CHAT_TIMESTAMP_FORMAT, time())..msg)
            else
                sinkTable[objectIcon..SPLITKEY..name..SPLITKEY..key]:Pour(msg)
            end
        end
    end
end

do-- VIGNETTE_ADDED
    local foundCooldowns = {}

    function MVAnnouncer:VIGNETTE_ADDED(event, vignetteGUID)
        local time_ticks = GetTime()
        if (foundCooldowns[vignetteGUID] or 0 ) <= time_ticks then
            foundCooldowns[vignetteGUID] = time_ticks + self.db.profile.cooldownTime

            local ofsX, ofsY, name, objectIcon = C_Vignettes_GetVignetteInfoFromInstanceID(vignetteGUID)
            self:Announce(name, objectIcon)
        end
    end
end

do-- WorldMap scanning
    local worldMap_waitTime = 60*5
    local foundCooldowns = {}

    function MVAnnouncer:WORLD_MAP_UPDATE() -- TODO: find better event when possible
        -- luacheck: globals NUM_WORLDMAP_POIS

        for i=1, NUM_WORLDMAP_POIS do
            local landmarkType, name, description, textureIndex, x, y, mapLinkID, inBattleMap, graveyardID, areaID, poiID, isObjectIcon, atlasIcon = GetMapLandmarkInfo(i)
            if landmarkType == LE_MAP_LANDMARK_TYPE_VIGNETTE then
                local time_ticks = GetTime()
                if ( foundCooldowns[name] or 0 ) <= time_ticks then
                    foundCooldowns[name] = time_ticks + worldMap_waitTime
                    self:Announce(name, textureIndex, true)
                    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
                end
            end
        end
    end

    function MVAnnouncer:ZONE_CHANGED_NEW_AREA()
        wipe(foundCooldowns)
        self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
    end
end

-- LOCALE: L["CONSOLE_HELP"] = "Chat commands:\n|cFF278BCB/mva|r - opens settings\n|cFF278BCB/mva|r manual - add a vignette manual to the settings\n|cFF278BCB/mva|r remove - remove a vignette from the settings"
-- LOCALE: L["CONSOLE_HELP_REMOVE"] = '|cFF278BCB/mva|r remove objecIconID objectName\nObjectName must be the same as in the option shown.\n\ne.g: 40 "Ohter Chest"'
function MVAnnouncer:CMD(input)
    local arg1, lastPos = self:GetArgs(input, 1)
    arg1 = arg1 and lower(arg1) or nil

    if not arg1 then
        self:OpenSettings()
    elseif arg1 == "manual" then
        self:CMD_Manual( sub(input, lastPos) )
    elseif arg1 ==  "remove" then
        local objectIcon, objectName
        objectIcon, objectName, lastPos = self:GetArgs(input, 2, lastPos)
        if tonumber(objectIcon) and type(objectName) == "string" and lastPos == 1e9 then
            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName] = nil
            self:RemoveVignette(objectIcon, objectName)
        else
            self:Print(L["CONSOLE_HELP_REMOVE"])
        end
    --[===[@debug@
    elseif arg1 == "debug" then
        local arg2 = self:GetArgs(input, 1, lastPos)
        local coords = OBJECTICONS_coords[arg2]
        if coords then
            local left, right, top, bottom = unpack(coords)
            left, right, top, bottom = left*objectIconTextureSize, right*objectIconTextureSize, top*objectIconTextureSize, bottom*objectIconTextureSize

            self:Printf("objectIcon; %s\nleft: %s, top: %s", arg2, left/32, top/32)
            self:Printf(OBJECTICONS_chatTexture, 16, 16, left, right, top, bottom)
        else
            self:Printf("No coords for %s", arg2 or "nil")
        end
    --@end-debug@]===]
    else
        self:Print(L["CONSOLE_HELP"])
    end
end

-- LOCALE: L["CONSOLE_HELP_MANUAL"] = '|cFF278BCB/mva|r manual objecIconID objectName active playSound\nObjectName must be the same as in the tooltip of the minimap vignette, use quotes for names with spacing.\n\nActive chest icon without sound e.g: 40 Chest true false\nInactive skull icon with default sound settings e.g: 41 Rar false\nChest icon with default active setting and active sound e.g: 40 "Other Chest" nil true'
function MVAnnouncer:CMD_Manual(input)
    local objectIcon, objectName, lastPos = self:GetArgs(input, 2)
    if tonumber(objectIcon) and type(objectName) == "string" and len(objectName) > 0 then
        if lastPos == 1e9 then
            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].seen = true
        else
            local isOK = true -- luacheck: ignore
            local active
            active, lastPos = self:GetArgs(input, 1, lastPos)

            active, isOK = GetBooleanValue(active)
            if active ~= nil and isOK then
                self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].active = active
            elseif not isOK then
                self:Print(L["CONSOLE_HELP_MANUAL"])
                return
            end

            if lastPos ~= 1e9 then
                local playSound
                playSound, lastPos = self:GetArgs(input, 1, lastPos)
                playSound, isOK = GetBooleanValue(playSound)
                if playSound ~= nil and isOK then
                    self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].playSound = playSound
                elseif not isOK then
                    self:Print(L["CONSOLE_HELP_MANUAL"])
                    return
                end
            end

            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].seen = true
        end
        self:NotifyOptionsChange()
    else
        self:Print(L["CONSOLE_HELP_MANUAL"])
    end
end

function MVAnnouncer:OpenSettings(...)
    InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    if not InterfaceOptionsFrameAddOns:IsShown() then
        InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    end
end


do-- options
    local options = {}
    local defaultsOptions = {}
    local announcementOptions = {}
    local outputMinChar, outputMaxChar = 3, 32

    local OptionsMethods = {}
    function OptionsMethods:Set(info, value, ...)
        MVAnnouncer.db.profile[info[#info]] = value
    end
    function OptionsMethods:Get(info, value, ...)
        return MVAnnouncer.db.profile[info[#info]]
    end
    function OptionsMethods:SetMapAnnounce(info, value, ...)
        self:Set(info, value, ...)
        if value then
            MVAnnouncer:RegisterEvent("WORLD_MAP_UPDATE")
        else
            MVAnnouncer:UnregisterEvent("WORLD_MAP_UPDATE")
        end
    end
    local tmpDefaults = {
        sinks = {}
    }
    local tmpDefaultsSinksTable = setmetatable({}, {__index=function(t,k)
        local v = rawget(t, k)

        if type(v) ~= "table" then
            v = LSK:Embed({})
            tmpDefaults.sinks[k] = tmpDefaults.sinks[k] or {}
            v:SetSinkStorage(tmpDefaults.sinks[k])
            rawset(t, k, v)
        end

        return v
    end})
    function OptionsMethods:SetDefaults(info, value, ...)
        tmpDefaults[info[#info]] = value
    end
    function OptionsMethods:GetDefaults(info, value, ...)
        if tmpDefaults[info[#info]] ~= nil then
            return tmpDefaults[info[#info]]
        else
            return MVAnnouncerDefaults[info[#info]]
        end
    end
    function OptionsMethods:ApplyDefaults(info, _value, ...)
        local db = MVAnnouncer.db
        db:RegisterDefaults(nil)

        for key, value in pairs(tmpDefaults) do
            if key == "sinks" then
                for sinkKey, sinkValue in pairs(value) do
                    if type(sinkValue) == "table" then
                        for settingKey, settingValue in pairs(sinkValue) do
                            if not MVAnnouncerDefaults[key][sinkKey] then
                                MVAnnouncerDefaults[key][sinkKey] = {}
                            end
                            MVAnnouncerDefaults[key][sinkKey][settingKey] = settingValue
                        end
                    else
                        MVAnnouncerDefaults[key][sinkKey] = nil
                    end
                end
            else
                MVAnnouncerDefaults[key] = value
            end
        end

        defaults.profile.objectIconTypeOfInterest['**']['**'] = MVAnnouncerDefaults
        db:RegisterDefaults(defaults)

        MVAnnouncer:CleanOptions()
    end
    function OptionsMethods:DeleteDefaultSink(info, value, ...)
        local sinkName = info.arg

        tmpDefaults.sinks[sinkName] = false
        tmpDefaultsSinksTable[sinkName] = nil
        defaultsOptions.args[sinkName.."_sink"] = nil
    end
    function OptionsMethods:ValidateDefaultOutputName(info, value, ...)
        value = trim(value)
        if match(value, "[%w]+[%w%s]*[%w]+") and len(value) >= outputMinChar and len(value) <= outputMaxChar then
            if not tmpDefaults.sinks[value] then
                return true
            end

            return L["Output %q exits already!"]:format(value)
        end

        return format(L["Output name length may be between %d and %d character and consist of letters, numbers, and spacings."], outputMinChar, outputMaxChar)
    end
    function OptionsMethods:CreateNewDefaultOutput(info, value, ...)
        value = trim(value)
        tmpDefaults.sinks[value] = tmpDefaults.sinks[value] or {}
        tmpDefaults.sinks[value].sink20OutputSink = "Default"
    end
    function OptionsMethods:SetAnnoucementsOptions(info, value, ...)
        local objectIcon = SPLITKEY:split(info.arg)

        MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][info[#info]] = value
    end
    function OptionsMethods:GetAnnoucementsOptions(info, value, ...)
        local objectIcon = split(SPLITKEY, info.arg)

        return MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][info[#info]]
    end
    function OptionsMethods:SetAnnoucementSettings(info, value, ...)
        local objectIcon, name = split(SPLITKEY, info.arg)

        MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name][info[#info]] = value
    end
    function OptionsMethods:GetAnnoucementSettings(info, value, ...)
        local objectIcon, name = split(SPLITKEY, info.arg)

        return MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name][info[#info]]
    end
    function OptionsMethods:ValidateOutputName(info, value, ...)
        value = trim(value)
        if match(value, "[%w]+[%w%s]*[%w]+") and len(value) >= outputMinChar and len(value) <= outputMaxChar then
            local objectIcon, name = split(SPLITKEY, info.arg)
            if MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[value].touched == nil then
                return true
            end

            return format(L["Output %q exits already!"], value)
        end

        return format(L["Output name length may be between %d and %d character and consist of letters, numbers, and spacings."], outputMinChar, outputMaxChar)
    end
    function OptionsMethods:CreateNewOutput(info, value, ...)
        value = trim(value)
        local objectIcon, name = split(SPLITKEY, info.arg)
        MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[value].touched = true
    end
    function OptionsMethods:GetNewOutput(info, value, ...)
        return ""
    end
    function OptionsMethods:DeleteSink(info, value, ...)
        local objectIcon, name, sinkName = split(SPLITKEY, info.arg)

        announcementOptions[objectIcon..name].args[sinkName] = nil
        MVAnnouncer.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[sub(sinkName, 1, -6)] = false
        sinkTable[info.arg] = nil
    end
    function OptionsMethods:SetColour(info, r, g, b, a, ...)
        self:SetAnnoucementSettings(info, RGBAToHex(r,g,b,a), ...)
    end
    function OptionsMethods:GetColour(info, value, ...)
        return HexToRGBA(self:GetAnnoucementSettings(info, value, ...))
    end
    function OptionsMethods:SetDefaultColour(info, r, g, b, a, ...)
        self:SetDefaults(info, RGBAToHex(r,g,b,a), ...)
    end
    function OptionsMethods:GetDefaultColour(info, value, ...)
        return HexToRGBA(self:GetDefaults(info, value, ...))
    end

    function MVAnnouncer.GetOptionsTable(uiType, uiName, appName)
        local self = MVAnnouncer
        options.name = ADDONNAME
        options.type = "group"
        options.childGroups = "tab"
        options.inline = true
        options.handler = OptionsMethods
        options.set = "Set"
        options.get = "Get"
        options.args = options.args or {}

        options.args.active = options.args.active or {}
        options.args.active.name = L["Active"]
        options.args.active.type = "toggle"
        options.args.active.desc = format(L["Enable/Disable %s on this profile."], ADDONNAME)
        options.args.active.order = 0

        options.args.announce_map_vignette = options.args.announce_map_vignette or {}
        options.args.announce_map_vignette.name = L["WorldMap Vignette"]
        options.args.announce_map_vignette.type = "toggle"
        options.args.announce_map_vignette.desc = L["Creates announcements of WorldMap Vignette"]
        options.args.announce_map_vignette.set = "SetMapAnnounce"
        options.args.announce_map_vignette.order = 0.5

        options.args.cooldownTime = options.args.cooldownTime or {}
        options.args.cooldownTime.name = L["Warning Cooldown"]
        options.args.cooldownTime.type = "range"
        options.args.cooldownTime.desc = format(L["How long should be wait before the same vignette can trigger a new announcement.\n%g Seconds."], self.db.profile.cooldownTime)
        options.args.cooldownTime.min = 0
        options.args.cooldownTime.max = 900
        options.args.cooldownTime.step = 0.01
        options.args.cooldownTime.bigStep = 1
        options.args.cooldownTime.softMax = 60
        options.args.cooldownTime.order = 0.55

        options.args.soundCooldown = options.args.soundCooldown or {}
        options.args.soundCooldown.name = L["Sound Cooldown"]
        options.args.soundCooldown.type = "range"
        options.args.soundCooldown.desc = L["Time until the same sound can be played again."]
        options.args.soundCooldown.min = 0
        options.args.soundCooldown.max = 900
        options.args.soundCooldown.step = 0.01
        options.args.soundCooldown.bigStep = 1
        options.args.soundCooldown.softMax = 10
        options.args.soundCooldown.order = 0.555



        for objectIcon, announcements in pairs(self.db.profile.objectIconTypeOfInterest) do
            options.args[objectIcon] = options.args[objectIcon] or {}

            if tonumber(objectIcon) then
                options.args[objectIcon].icon = OBJECTICONS_texture
                options.args[objectIcon].iconCoords = OBJECTICONS_coords[objectIcon]
                options.args[objectIcon].name = L["Vignette: %s"]:format(objectIcon)
            else
                options.args[objectIcon].name = objectIcon
            end

            options.args[objectIcon].type = "group"
            options.args[objectIcon].childGroups = "tree"
            options.args[objectIcon].inline = false
            options.args[objectIcon].set = "SetAnnoucementsOptions"
            options.args[objectIcon].get = "GetAnnoucementsOptions"
            options.args[objectIcon].args = options.args[objectIcon].args or {}

            if tonumber(objectIcon) then
                options.args[objectIcon].args.image = options.args[objectIcon].args.image or {}
                options.args[objectIcon].args.image.name = L["-  \"Vignette\""]
                options.args[objectIcon].args.image.type = "description"
                options.args[objectIcon].args.image.image = OBJECTICONS_texture
                options.args[objectIcon].args.image.imageCoords = OBJECTICONS_coords[objectIcon]
                options.args[objectIcon].args.image.fontSize = "medium"
                options.args[objectIcon].args.image.order = 0
            end

            options.args[objectIcon].args.active = options.args[objectIcon].args.active or {}
            options.args[objectIcon].args.active.name = L["Active"]
            options.args[objectIcon].args.active.desc = L["If this vignette type active."]
            options.args[objectIcon].args.active.arg = objectIcon
            options.args[objectIcon].args.active.type = "toggle"
            options.args[objectIcon].args.active.order = 0.1

            for name, value in pairs(announcements) do
                if type(value) == "table" then
                    options.args[objectIcon].args[name] = self.GetAnnoucementSettings(uiType, uiName, appName, objectIcon, name)
                end
            end
        end

        return options
    end

    function MVAnnouncer.GetAnnoucementSettings(uiType, uiName, appName, objectIcon, name)
        local self = MVAnnouncer

        announcementOptions[objectIcon..name] = announcementOptions[objectIcon..name] or {}
        local settings = announcementOptions[objectIcon..name]
        local arg = objectIcon..SPLITKEY..name

        settings.name = name
        settings.type = "group"
        settings.set = "SetAnnoucementSettings"
        settings.get = "GetAnnoucementSettings"
        settings.args = settings.args or {}

        settings.args.active = settings.args.active or {}
        settings.args.active.name = L["Active"]
        settings.args.active.type = "toggle"
        settings.args.active.desc = L["If this announcement active."]
        settings.args.active.arg = arg
        settings.args.active.width = "double"
        settings.args.active.order = 0

        settings.args.iconSize = settings.args.iconSize or {}
        settings.args.iconSize.name = L["Icon size"]
        settings.args.iconSize.type = "range"
        settings.args.iconSize.desc = L["Size of the vignette icon."]
        settings.args.iconSize.arg = arg
        settings.args.iconSize.min = 8
        settings.args.iconSize.max = 512
        settings.args.iconSize.step = 1
        settings.args.iconSize.bigStep = 4
        settings.args.iconSize.softMax = 64
        settings.args.iconSize.order = 2

        settings.args.showIconBefore = settings.args.showIconBefore or {}
        settings.args.showIconBefore.name = L["Show an icon before the announcement"]
        settings.args.showIconBefore.type = "toggle"
        settings.args.showIconBefore.desc = L["Should be shown an icon before the announcement."]
        settings.args.showIconBefore.arg = arg
        settings.args.showIconBefore.width = "full"
        settings.args.showIconBefore.order = 3.1

        settings.args.showIconAfter = settings.args.showIconAfter or {}
        settings.args.showIconAfter.name = L["Show an icon after the announcement"]
        settings.args.showIconAfter.type = "toggle"
        settings.args.showIconAfter.desc = L["Should be shown an icon after the announcement."]
        settings.args.showIconAfter.arg = arg
        settings.args.showIconAfter.width = "full"
        settings.args.showIconAfter.order = 3.2

        settings.args.mainTextColour = settings.args.mainTextColour or {}
        settings.args.mainTextColour.name = L["Main text colour"]
        settings.args.mainTextColour.type = "color"
        settings.args.mainTextColour.desc = L["Main text colour of the announcement"]
        settings.args.mainTextColour.hasAlpha = true
        settings.args.mainTextColour.arg = arg
        settings.args.mainTextColour.set = "SetColour"
        settings.args.mainTextColour.get = "GetColour"
        settings.args.mainTextColour.width = "full"
        settings.args.mainTextColour.order = 4.1

        settings.args.objectTextColour = settings.args.objectTextColour or {}
        settings.args.objectTextColour.name = L["Vignette text colour"]
        settings.args.objectTextColour.type = "color"
        settings.args.objectTextColour.desc = L["Vignette text colour of the announcement"]
        settings.args.objectTextColour.hasAlpha = true
        settings.args.objectTextColour.arg = arg
        settings.args.objectTextColour.set = "SetColour"
        settings.args.objectTextColour.get = "GetColour"
        settings.args.objectTextColour.width = "full"
        settings.args.objectTextColour.order = 4.2

        settings.args.playSound = settings.args.playSound or {}
        settings.args.playSound.name = L["Play sound"]
        settings.args.playSound.type = "toggle"
        settings.args.playSound.desc = L["Should this announcement play a sound."]
        settings.args.playSound.arg = arg
        settings.args.playSound.order = 5.1

        settings.args.sound = settings.args.sound or {}
        settings.args.sound.name = L["Sound"]
        settings.args.sound.type = "select"
        settings.args.sound.desc = L["Which sound should be played."]
        settings.args.sound.arg = arg
        settings.args.sound.values = LSM:HashTable(LSM.MediaType.SOUND)
        settings.args.sound.style = "dropdown"
        settings.args.sound.dialogControl = "LSM30_Sound"
        settings.args.sound.width = "double"
        settings.args.sound.order = 5.2

        settings.args.newOutput = settings.args.newOutput or {}
        settings.args.newOutput.name = L["Create new output"]
        settings.args.newOutput.type = "input"
        settings.args.newOutput.desc = L["Add an other output. Name may only consist letters, numbers, and spacings."]
        settings.args.newOutput.arg = arg
        settings.args.newOutput.validate = "ValidateOutputName"
        settings.args.newOutput.set = "CreateNewOutput"
        settings.args.newOutput.get = "GetNewOutput"
        settings.args.newOutput.order = 6

        local i = 1
        for key, value in pairs(self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks) do
            if value ~= false then

                local isDefault = not self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[key].touched

                settings.args[key.."_sink"] = sinkTable[arg..SPLITKEY..key]:GetSinkAce3OptionsDataTable()
                if key == "BuildInDefaultOutputSink" then
                    settings.args[key.."_sink"].name = L["Default output"]
                else
                    settings.args[key.."_sink"].name = key
                end

                settings.args[key.."_sink"].order = 10+i/10
                settings.args[key.."_sink"].plugins = settings.args[key.."_sink"].plugins or {}
                settings.args[key.."_sink"].plugins[ADDONNAME] = settings.args[key.."_sink"].plugins[ADDONNAME] or {}
                settings.args[key.."_sink"].plugins[ADDONNAME].delete = settings.args[key.."_sink"].plugins[ADDONNAME].delete or {}
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.name = L["Delete output"]
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.type = "execute"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.desc = L["Click here to delete this output."]
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.disabled = isDefault
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.arg = arg..SPLITKEY..key.."_sink"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.confirm = true
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.confirmText = format(L["Delete %q output"], settings.args[key.."_sink"].name)
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.func = "DeleteSink"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.order = -0.1

                if isDefault then
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault = settings.args[key.."_sink"].plugins[ADDONNAME].isDefault or {}
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.name = L["|cFF999999You can't delete a default output.|r"]
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.type = "description"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.fontSize = "small"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.width = "full"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.order = -0.01
                end

                i = i + 1
            end
        end

        return settings
    end

    function MVAnnouncer.GetDefaultsOptionTable(uiType, uiName, appName)
        defaultsOptions = defaultsOptions or {}
        defaultsOptions.name = L["Default Settings"]
        defaultsOptions.type = "group"
        defaultsOptions.desc = L["Default Settings for unchanged announcements."]
        defaultsOptions.handler = OptionsMethods
        defaultsOptions.set = "SetDefaults"
        defaultsOptions.get = "GetDefaults"
        defaultsOptions.args = defaultsOptions.args or {}

        defaultsOptions.args.active = defaultsOptions.args.active or {}
        defaultsOptions.args.active.name = L["Active"]
        defaultsOptions.args.active.type = "toggle"
        defaultsOptions.args.active.desc = L["If a new announcement should be active."]
        defaultsOptions.args.active.width = "double"
        defaultsOptions.args.active.order = 0

        defaultsOptions.args.showIconBefore = defaultsOptions.args.showIconBefore or {}
        defaultsOptions.args.showIconBefore.name = L["Show an icon before the announcement"]
        defaultsOptions.args.showIconBefore.type = "toggle"
        defaultsOptions.args.showIconBefore.desc = L["If a new announcement should show an icon before the announcement."]
        defaultsOptions.args.showIconBefore.width = "full"
        defaultsOptions.args.showIconBefore.order = 3.1

        defaultsOptions.args.showIconAfter = defaultsOptions.args.showIconAfter or {}
        defaultsOptions.args.showIconAfter.name = L["Show an icon after the announcement"]
        defaultsOptions.args.showIconAfter.type = "toggle"
        defaultsOptions.args.showIconAfter.desc = L["If a new announcement should show an icon after the announcement."]
        defaultsOptions.args.showIconAfter.width = "full"
        defaultsOptions.args.showIconAfter.order = 3.2

        defaultsOptions.args.iconSize = defaultsOptions.args.iconSize or {}
        defaultsOptions.args.iconSize.name = L["Icon size"]
        defaultsOptions.args.iconSize.type = "range"
        defaultsOptions.args.iconSize.desc = L["Size of the icon from a new announcement."]
        defaultsOptions.args.iconSize.min = 8
        defaultsOptions.args.iconSize.max = 512
        defaultsOptions.args.iconSize.step = 1
        defaultsOptions.args.iconSize.bigStep = 4
        defaultsOptions.args.iconSize.softMax = 64
        defaultsOptions.args.iconSize.order = 2

        defaultsOptions.args.mainTextColour = defaultsOptions.args.mainTextColour or {}
        defaultsOptions.args.mainTextColour.name = L["Main text colour"]
        defaultsOptions.args.mainTextColour.type = "color"
        defaultsOptions.args.mainTextColour.desc = L["Main text colour of a new announcement"]
        defaultsOptions.args.mainTextColour.hasAlpha = true
        defaultsOptions.args.mainTextColour.set = "SetDefaultColour"
        defaultsOptions.args.mainTextColour.get = "GetDefaultColour"
        defaultsOptions.args.mainTextColour.width = "full"
        defaultsOptions.args.mainTextColour.order = 4.1

        defaultsOptions.args.objectTextColour = defaultsOptions.args.objectTextColour or {}
        defaultsOptions.args.objectTextColour.name = L["Vignette text colour"]
        defaultsOptions.args.objectTextColour.type = "color"
        defaultsOptions.args.objectTextColour.desc = L["Vignette text colour of a new announcement"]
        defaultsOptions.args.objectTextColour.hasAlpha = true
        defaultsOptions.args.objectTextColour.set = "SetDefaultColour"
        defaultsOptions.args.objectTextColour.get = "GetDefaultColour"
        defaultsOptions.args.objectTextColour.width = "full"
        defaultsOptions.args.objectTextColour.order = 4.2

        defaultsOptions.args.playSound = defaultsOptions.args.playSound or {}
        defaultsOptions.args.playSound.name = L["Play sound"]
        defaultsOptions.args.playSound.type = "toggle"
        defaultsOptions.args.playSound.desc = L["If a new announcement should play a sound."]
        defaultsOptions.args.playSound.order = 5.1

        defaultsOptions.args.sound = defaultsOptions.args.sound or {}
        defaultsOptions.args.sound.name = L["Sound"]
        defaultsOptions.args.sound.type = "select"
        defaultsOptions.args.sound.desc = L["Which sound should be played."]
        defaultsOptions.args.sound.values = LSM:HashTable(LSM.MediaType.SOUND)
        defaultsOptions.args.sound.style = "dropdown"
        defaultsOptions.args.sound.dialogControl = "LSM30_Sound"
        defaultsOptions.args.sound.width = "double"
        defaultsOptions.args.sound.order = 5.2

        defaultsOptions.args.newOutput = defaultsOptions.args.newOutput or {}
        defaultsOptions.args.newOutput.name = L["Create new default output"]
        defaultsOptions.args.newOutput.type = "input"
        defaultsOptions.args.newOutput.desc = L["Add an other default output. Name may only consist letters, numbers, and spacings."]
        defaultsOptions.args.newOutput.validate = "ValidateDefaultOutputName"
        defaultsOptions.args.newOutput.set = "CreateNewDefaultOutput"
        defaultsOptions.args.newOutput.get = "GetNewOutput"
        defaultsOptions.args.newOutput.order = 6


        for key in pairs(MVAnnouncerDefaults.sinks) do
            if tmpDefaults.sinks[key] == nil then
                tmpDefaults.sinks[key] = {}
                for sinkKey, value in pairs(MVAnnouncerDefaults.sinks[key]) do
                    tmpDefaults.sinks[key][sinkKey] = value
                end
            end
        end
        local i = 1
        for key, value in pairs(tmpDefaults.sinks) do
            if value ~= false then
                defaultsOptions.args[key.."_sink"] = tmpDefaultsSinksTable[key]:GetSinkAce3OptionsDataTable()

                if key ~= '**' then
                    if key == "BuildInDefaultOutputSink" then
                        defaultsOptions.args[key.."_sink"].name = L["Default output"]
                    else
                        defaultsOptions.args[key.."_sink"].name = key
                    end

                    defaultsOptions.args[key.."_sink"].plugins = defaultsOptions.args[key.."_sink"].plugins or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.name = L["Delete output"]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.type = "execute"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.desc = L["Click here to delete this output."]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.arg = key
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.confirm = true
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.confirmText = format(L["Delete %q output"], defaultsOptions.args[key.."_sink"].name)
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.func = "DeleteDefaultSink"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.order = -0.1
                else
                    defaultsOptions.args[key.."_sink"].name = L["Standard"]
                    defaultsOptions.args[key.."_sink"].desc = L["This are the settings for all not changed and new created outputs."]
                    defaultsOptions.args[key.."_sink"].plugins = defaultsOptions.args[key.."_sink"].plugins or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.name = L["This are the settings for all not changed and new created outputs."]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.type = "description"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.fontSize = "medium"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.order = 1
                end

                defaultsOptions.args[key.."_sink"].order = 10+i/10
                i = i + 1
            end
        end

        defaultsOptions.args.applyDefaults = defaultsOptions.args.applyDefaults or {}
        defaultsOptions.args.applyDefaults.name = L["Apply defaults"]
        defaultsOptions.args.applyDefaults.type = "execute"
        defaultsOptions.args.applyDefaults.desc = L["Set this settings as defaults."]
        defaultsOptions.args.applyDefaults.func = "ApplyDefaults"
        defaultsOptions.args.applyDefaults.order = 0

        return defaultsOptions
    end

    function MVAnnouncer:CleanOptions()
        wipe(announcementOptions)
        wipe(options)
        wipe(sinkTable)
        self:NotifyOptionsChange()
    end

    function MVAnnouncer:RemoveVignette(objectIcon, objectName)
        if type(options.args[objectIcon]) == "table" then
            options.args[objectIcon].args[objectName] = nil
        end
        self:NotifyOptionsChange()
    end

    function MVAnnouncer:NotifyOptionsChange()
        LibStub("AceConfigRegistry-3.0"):NotifyChange(ADDONNAME)
    end

    function MVAnnouncer:RegisterOptions()
        local AceConfig = LibStub("AceConfig-3.0")
        local AceConfigDialog = LibStub("AceConfigDialog-3.0")

        AceConfig:RegisterOptionsTable(ADDONNAME, self.GetOptionsTable )
        AceConfig:RegisterOptionsTable(ADDONNAME.."defaults", self.GetDefaultsOptionTable )
        AceConfig:RegisterOptionsTable(ADDONNAME.."profile", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db) )

        AceConfigDialog:AddToBlizOptions(ADDONNAME, ADDONNAME)
        AceConfigDialog:AddToBlizOptions(ADDONNAME.."defaults", L["Defaults"], ADDONNAME)
        AceConfigDialog:AddToBlizOptions(ADDONNAME.."profile", L["Profile"], ADDONNAME)
    end
end







